#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <unistd.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/wait.h>

static const char *progname = "timeit";

static void set_progname(const char *prog)
{
  const char *p;

  p = strrchr(prog, '/');
  progname = p ? p + 1 : progname;
}

static void lose(const char *msg, ...)
{
  va_list ap;

  va_start(ap, msg);
  fprintf(stderr, "%s: ", progname);
  vfprintf(stderr, msg, ap);
  fputc('\n', stderr);
  va_end(ap);
  exit(127);
}

static double timeval_to_float(const struct timeval *tv)
  { return (tv->tv_sec + tv->tv_usec*1e-6); }

int main(int argc, char *argv[])
{
  struct rusage ru;
  struct timeval t0, t1, t;
  pid_t kid;
  int i, st;

  set_progname(argv[0]);
  gettimeofday(&t0, 0);
  kid = fork(); if (kid < 0) lose("fork failed: %s", strerror(errno));
  if (!kid) {
    execvp(argv[1], argv + 1);
    lose("exec (`%s') failed: %s", argv[1], strerror(errno));
  }
  if (wait4(kid, &st, 0, &ru) < 0) lose("wait failed: %s", strerror(errno));
  gettimeofday(&t1, 0);
  if (st) {
    if (WIFSIGNALED(st))
      lose("program killed by signal %d\n", WTERMSIG(st));
    else if (WIFEXITED(st))
      lose("program failed with status %d\n", WEXITSTATUS(st));
    else
      lose("program exited with incomprehensible status 0x%04x\n", st);
  }

  if (t0.tv_usec > t1.tv_usec) {
    t.tv_sec = t1.tv_sec - t0.tv_sec - 1;
    t.tv_usec = t1.tv_usec + 1000000 - t0.tv_usec;
  } else {
    t.tv_sec = t1.tv_sec - t0.tv_sec;
    t.tv_usec = t1.tv_usec - t0.tv_usec;
  }

  for (i = 1; i < argc; i++) {
    if (i > 1) fputc(' ', stderr);
    fputs(argv[i], stderr);
  }
  fprintf(stderr, ": elapsed = %.4fs; user = %.4fs; system = %.4fs\n",
	  timeval_to_float(&t),
	  timeval_to_float(&ru.ru_utime),
	  timeval_to_float(&ru.ru_stime));
  return (0);
}
